#!/usr/bin/env python
# coding: utf-8

from __future__ import annotations
from typing import Literal, Sequence, Mapping
from dataclasses import dataclass, field
from math import log10, ceil
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.colors import rgb2hex
from matplotlib.ticker import MaxNLocator
from matplotlib import transforms as mtransforms
from ._type_alias import Color, Point, Axes


# Type alias
BasePairUnit = Literal["bp", "kb", "Mb", "Gb", "Tb"]


def get_cmap_colors(
    cmap_name: str, format_: Literal["hex", "rgb"] = "hex"
) -> list[Color]:
    """
    Get all colors for a given matplotlib palette.

    Adapted from https://gist.github.com/jdbcode/33d37999f950a36b43e058d15280b536.

    >>> get_cmap_colors("Set2")
    ['#66c2a5', '#fc8d62', '#8da0cb', '#e78ac3', '#a6d854', '#ffd92f', '#e5c494', '#b3b3b3']
    """
    cmap = plt.get_cmap(cmap_name)
    colors = list({cmap(i)[:3]: None for i in range(cmap.N)})
    if format_ == "rgb":
        pass
    elif format_ == "hex":
        colors = [rgb2hex(c) for c in colors]
    else:
        raise ValueError(
            f"Invalid value for `format_`: {format_}. Expecting one of {'rgb', 'hex'}"
        )
    return colors


def get_random_colors(n: int, *, seed=0, cmap="hsv") -> list[Color]:
    """
    Returns ``n`` random colors selected uniformly from ``cmap``.

    >>> get_random_colors(3)
    [(0.0, 0.22463448382566054, 1.0, 1.0), (0.4018366371307548, 1.0, 0.0, 1.0), (1.0, 0.2316178786767022, 0.0, 1.0)]
    """
    rng = np.random.default_rng(seed=0)
    cmap = plt.get_cmap("hsv")
    colors = [cmap(rng.random()) for __ in range(n)]
    return colors


def draw_rigid_polygon(
    ax: Axes,
    shape: Sequence[Point],
    position: Point,
    position_transform: mpl.transforms.Transform,
    **kw,
):
    """
    Draw a polygon patch with shape defined in inches and position defined in data, Axes or physical units.
    """
    if (
        len(position) != 2
        or not isinstance(position[0], float)
        or not isinstance(position[1], float)
    ):
        raise ValueError(f"position={position}")
    transform = ax.figure.dpi_scale_trans + mpl.transforms.ScaledTranslation(
        *position, position_transform
    )
    polygon = mpl.patches.Polygon(
        shape,
        transform=transform,
        **kw,
    )
    ax.add_patch(polygon)


def get_ax_size(ax: Axes):
    """
    Return the size of a given Axes in inches.

    >>> fig, ax = plt.subplots(figsize=(8, 6))
    >>> get_ax_size(ax)
    (6.2, 4.62)
    """
    fig = ax.figure
    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    width, height = bbox.width, bbox.height
    return (width, height)


def scientific_notation(
    x: float, significant_figures: int = 3, *, quote: str = "$"
) -> str:
    r"""
    Returns scientific notation in Matplotlib mathtext format.

    >>> scientific_notation(0.000000013923, 4)
    '$1.392\\times 10^{-8}$'
    """
    if significant_figures < 1:
        raise ValueError()
    if not isinstance(significant_figures, int):
        raise ValueError()
    float_digits = significant_figures - 1
    # Python native scientific notation
    coefficient, exponent = format(x, f".{float_digits}e").split("e")
    # Remove leading zeros
    exponent = str(int(exponent))
    # Format with mathtext
    s = quote + coefficient + r"\times 10^" + "{" + exponent + "}" + quote
    return s


class PrunedMaxNLocator(MaxNLocator):
    """
    A Matplotlib `tick locator <https://matplotlib.org/stable/api/ticker_api.html#tick-locating>`_ that prunes ticks near the edge of the plot.
    """

    def __init__(self, nbins: int = 4, prune: tuple[float, float] = (0.05, 0.05)):
        """
        :param nbins: Maximum number of intervals; one less than max number of ticks.
        :param prune: A two-element :py:class:`tuple` `(prune_lower, prune_upper)`. Ticks outside the interval [vmin + prune_lower × (vmax - vmin), vmax - self._prune_upper × (vmax - vmin)] will be pruned. See :py:meth:`.PrunedMaxNLocator.tick_values`.
        """
        prune_lower, prune_upper = prune
        if not (0 <= prune_lower <= 1 and 0 <= prune_upper <= 1):
            raise ValueError("`prune` must be a pair of numbers between 0 and 1.")
        self._prune_lower, self._prune_upper = prune_lower, prune_upper

        super().__init__(nbins=nbins, prune=None)

    def tick_values(self, vmin: float, vmax: float):
        if self._symmetric:
            vmax = max(abs(vmin), abs(vmax))
            vmin = -vmax
        vmin, vmax = mtransforms.nonsingular(vmin, vmax, expander=1e-13, tiny=1e-14)
        locs = self._raw_ticks(vmin, vmax)
        min_tick_location = vmin + self._prune_lower * (vmax - vmin)
        max_tick_location = vmax - self._prune_upper * (vmax - vmin)
        locs = [x for x in locs if min_tick_location <= x <= max_tick_location]
        return self.raise_if_exceeds(locs)


@dataclass(init=False)
class BasePairFormatter(mpl.ticker.FuncFormatter):
    """
    A Matplotlib `tick formatter <https://matplotlib.org/stable/api/ticker_api.html#tick-formatting>`_ for common base pair units (bp, kb, Mb, Gb, Tb).

    >>> formatter = BasePairFormatter('kb')
    >>> formatter(24310)
    '24.310 kb'
    >>> formatter = BasePairFormatter('Mb', 2, show_suffix=False)
    >>> formatter(844_293_192)
    '844.29'
    """

    unit: BasePairUnit
    decimals: int
    show_suffix: bool

    UNITS: Sequence[BasePairUnit] = field(
        default=("bp", "kb", "Mb", "Gb", "Tb"), repr=False
    )
    DIVISORS: Sequence[int] = field(
        default=(1, int(1e3), int(1e6), int(1e9), int(1e12)), repr=False
    )

    def __init__(
        self,
        unit: BasePairUnit,
        decimals: int | None = None,
        *,
        show_suffix: bool = True,
    ):
        """
        :param unit: Base pair unit.
        :param decimals: The number of decimal places to show for the coefficient. The default is ``0`` for ``'bp'``, or ``3`` for other values of ``unit``.
        :param show_suffix: Whether to show the unit as a suffix.
        """
        self.unit = unit
        self.show_suffix = show_suffix
        unit_divisor = self._get_unit_divisor(unit)

        if decimals is None:
            if unit == "bp":
                decimals = 0
            else:
                decimals = 3
        self.decimals = decimals

        def formatter_function(x: float, pos) -> str:
            tick_label = format(x / unit_divisor, f".{decimals}f")
            # print(unit_divisor)
            if show_suffix:
                tick_label += " " + unit
            return tick_label

        super().__init__(formatter_function)

    @staticmethod
    def _get_unit_divisor(unit: BasePairUnit) -> int:
        UNITS = BasePairFormatter.UNITS
        for _unit, divisor in zip(BasePairFormatter.UNITS, BasePairFormatter.DIVISORS):
            if _unit == unit:
                return divisor
        else:
            raise ValueError(
                f"Invalid value for unit: {unit!r}. Supported values: {tuple(UNITS)!r}."
            )

    @classmethod
    def from_limits(cls, limits: tuple[float, float], *, show_suffix: bool = True):
        """
        Automatically select reasonable `unit` and `decimals` based on axes limits.

        >>> fig, ax = plt.subplots()
        >>> ax.set_xlim(3200000, 5840000)
        (3200000.0, 5840000.0)
        >>> formatter = BasePairFormatter.from_limits(ax.get_xlim())
        >>> formatter
        BasePairFormatter(unit='Mb', decimals=1, show_suffix=True)
        >>> ax.xaxis.set_major_formatter(formatter)
        """

        start, end = limits
        if start > end:
            start, end = end, start
        span = end - start
        for unit, unit_divisor in zip(reversed(cls.UNITS), reversed(cls.DIVISORS)):
            if unit_divisor <= start:
                break
        decimals: int = max(0, 2 - ceil(log10(span / unit_divisor)))
        formatter = cls(
            unit,
            decimals,
            show_suffix=show_suffix,
        )
        return formatter



#!/usr/bin/env python
# coding: utf-8

from __future__ import annotations
from typing import Optional, Sequence, Literal, cast
from dataclasses import dataclass
from math import log10, ceil
from warnings import warn
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.transforms import Bbox
from IPython.display import display
import ipywidgets
from ._type_alias import Figure, Axes
from .plot import BasePairFormatter, PrunedMaxNLocator


class _GenomeViewerWidget:
    def __init__(self, figure: Figure, axes: Sequence[Axes]):
        self.figure: Figure = figure
        self.axes: Axes = axes
        self._initial_xlim: tuple[float, float] = self.get_xlim()

        center_widget = ipywidgets.Output()

        zoom_in_button = ipywidgets.Button(description="Zoom in")
        zoom_out_button = ipywidgets.Button(description="Zoom out")
        zoom_in_button.on_click(lambda button: self._zoom(1 / 1.2))
        zoom_out_button.on_click(lambda button: self._zoom(1.2))

        shift_left_button = ipywidgets.Button(description="<<")
        shift_right_button = ipywidgets.Button(description=">>")
        shift_left_button.on_click(lambda __: self._shift(-0.15))
        shift_right_button.on_click(lambda __: self._shift(0.15))

        reset_button = ipywidgets.Button(description="Reset")
        reset_button.on_click(lambda _: self._reset_xlim())

        region_text = ipywidgets.Text(
            value="",
            placeholder="",
            description="",
            disabled=False,
        )
        self._region_text = region_text
        go_button = ipywidgets.Button(description="Go")
        go_button.on_click(lambda __: self._goto_region(self._region_text.value))

        footer_widget = ipywidgets.VBox(
            [
                ipywidgets.HBox([region_text, go_button, reset_button]),
                ipywidgets.HBox(
                    [
                        shift_left_button,
                        shift_right_button,
                        zoom_in_button,
                        zoom_out_button,
                    ]
                ),
            ]
        )
        self._center_widget = center_widget
        self._footer_widget = footer_widget

        self._app = ipywidgets.AppLayout(
            center=center_widget, footer=footer_widget, pane_heights=[0, 1, "100px"]
        )

    @property
    def app(self) -> ipywidgets.AppLayout:
        return self._app

    def _ipython_display_(self) -> None:
        self._initial_xlim = self.get_xlim()
        self._update_display()
        display(self.app)

    def get_xlim(self) -> tuple[float, float]:
        return self.axes[-1].get_xlim()

    def set_xlim(self, *args, **kw) -> tuple[float, float]:
        return self.axes[-1].set_xlim(*args, **kw)

    def _zoom(self, scale: float, /) -> None:
        for ax in self.figure.axes:
            old_xlim = ax.get_xlim()
            center = sum(old_xlim) / 2
            radius = (old_xlim[1] - old_xlim[0]) / 2
            new_radius = radius * scale
            new_xlim = (center - new_radius, center + new_radius)
            ax.set_xlim(new_xlim)
        self._update_display()

    def _shift(self, distance: float, /) -> None:
        for ax in self.figure.axes:
            old_xlim = ax.get_xlim()
            radius = (old_xlim[1] - old_xlim[0]) / 2
            new_xlim = (
                old_xlim[0] + distance * radius,
                old_xlim[1] + distance * radius,
            )
            ax.set_xlim(new_xlim)
        self._update_display()

    def _goto(self, start: float, end: float) -> None:
        self.set_xlim(start, end)
        self._update_display()

    def _goto_region(self, region: str) -> None:
        start, end = region.replace(" ", "").replace(",", "").split("-")
        if start.isnumeric() and end.isnumeric():
            self._goto(float(start), float(end))
        else:
            raise ValueError(f"Invalid region {region!r}")

    def _reset_xlim(self) -> None:
        self.set_xlim(self._initial_xlim)
        self._update_display()

    def _update_region_text(self) -> None:
        start, end = self.get_xlim()
        self._region_text.value = f"{int(start):,} - {int(end):,}"

    def _update_display(self):
        self._update_region_text()
        self._center_widget.clear_output(wait=True)
        with self._center_widget:
            display(self.figure)


class GenomeViewer:
    """
    An interactive widget for viewing genomic data in Jupyter Notebooks.
    """

    def __init__(
        self,
        tracks: int = 1,
        *,
        figsize: tuple[float, float] = (8, 8),
        height_ratios: Optional[Sequence[float]] = None,
        constrained_layout: bool = True,
        figure_kw=dict(),
    ) -> None:

        with plt.ioff():
            fig, axes = plt.subplots(
                nrows=tracks,
                ncols=1,
                figsize=figsize,
                sharex=True,
                sharey=False,
                squeeze=False,
                gridspec_kw=dict(height_ratios=height_ratios),
                constrained_layout=constrained_layout,
                **figure_kw,
            )
            horizontal_padding = 0.2
            fig.get_layout_engine().set(w_pad=horizontal_padding)

            # Create a hidden Axes that is as wide as the Figure and has no height
            # This prevents the Figure from resizing when changing xlim interactively
            shadow_ax = fig.add_axes(Bbox.from_extents(0, 0.5, 1, 0.5))
            shadow_ax.axis("off")

        self._figure = fig
        self._axes = axes[:, 0]
        self._widget: _GenomeViewerWidget | None = None
        self.axes[-1].xaxis.set_major_locator(PrunedMaxNLocator(4, prune=(0.05, 0.05)))
        self.axes[-1].xaxis.set_major_formatter(BasePairFormatter('bp', 0, show_suffix=False))

    @property
    def figure(self) -> Figure:
        return self._figure

    @property
    def axes(self) -> Sequence[Axes]:
        return self._axes

    @property
    def widget(self) -> _GenomeViewerWidget:
        if self._widget is None:
            widget = _GenomeViewerWidget(self.figure, self.axes)
            self._widget = widget
        return self._widget

    @property
    def xaxis(self) -> mpl.axis.XAxis:
        return self.axes[-1].xaxis

    def set_xlim(self, *args, **kw) -> tuple[float, float]:
        start, end = self.axes[-1].set_xlim(*args, **kw)
        return start, end

    def get_xlim(self) -> tuple[float, float]:
        return self.axes[-1].get_xlim()

    def set_major_formatter(self, formatter: mpl.ticker.Formatter) -> None:
        self.axes[-1].xaxis.set_major_formatter(formatter)

    def set_minor_formatter(self, formatter: mpl.ticker.Formatter) -> None:
        self.axes[-1].xaxis.set_minor_formatter(formatter)

    def set_xlabel(self, *args, **kw) -> mpl.text.Text:
        self.axes[-1].set_xlabel(*args, **kw)

    def set_title(self, *args, **kw) -> mpl.text.Text:
        self.axes[0].set_title(*args, **kw)

    def savefig(self, *args, dpi=300, bbox_inches="tight", **kw):
        return self.figure.savefig(*args, dpi=300, bbox_inches=bbox_inches, **kw)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(figure={self.figure!r})"

    def clear(self) -> None:
        """
        Clear all contents in each Axes.
        """
        for ax in self.axes:
            ax.cla()



#!/usr/bin/env python
# coding: utf-8

from __future__ import annotations
import tempfile
import os
import pathlib
from typing import Literal, Callable, Optional, Union
import pysam
from .region_notation import (
    parse_region_notation,
    normalize_region_notation,
    get_region_notation,
)


def download_bam(
    bam_url: str,
    output_bam_path: str,
    region: str | tuple[str, tuple[int, int]] | tuple[str, None] | None = None,
    *,
    index_url: Optional[str] = None,
    output_index_path: Optional[str] = None,
    index: bool = True,
    override: bool = False,
) -> None:
    if index_url is None:
        index_url = bam_url + ".bai"
    if output_index_path is None:
        output_index_path = output_bam_path + ".bai"
    region_notation: Optional[str]
    if region is None:
        region_notation = None
    elif isinstance(region, str):
        region_notation = normalize_region_notation(region)
    elif isinstance(region, tuple):
        reference_name, interval = region
        region_notation = get_region_notation(reference_name, interval)
    else:
        raise TypeError(
            f"Invalid type for `region`: {type(region)!r}. Expecting str | tuple[str, tuple[int, int]] | tuple[str, None] | None."
        )

    if not os.path.isfile(output_bam_path) or override:
        workdir = os.getcwd()
        with tempfile.TemporaryDirectory() as d:
            try:
                os.chdir(d)
                params = ["-X", "-b", bam_url, index_url]
                if region_notation:
                    params.append(region_notation)
                bam_data = pysam.view(*params)  # type: ignore # pysam.view is not correctly typed.
            finally:
                os.chdir(workdir)
        target_dir = os.path.split(output_bam_path)[0]
        pathlib.Path(target_dir).mkdir(parents=True, exist_ok=True)
        with open(output_bam_path, "wb") as f:
            f.write(bam_data)
    if index and (not os.path.isfile(output_index_path) or override):
        target_dir = os.path.split(output_index_path)[0]
        pathlib.Path(target_dir).mkdir(parents=True, exist_ok=True)
        pysam.index(output_bam_path, output_index_path)  # type: ignore # pysam.index is not correctly typed.



#!/usr/bin/env python
# coding: utf-8

from __future__ import annotations
from typing import Literal, Callable, Any, Optional, TypeVar
from collections.abc import Iterable, Sequence
import os
import pathlib
import tempfile
from numbers import Real
import matplotlib as mpl
from matplotlib.colors import rgb2hex
import pysam
from ._type_alias import Color, Identifier

T = TypeVar("T")

def key_sort(sequence: Sequence[T], keys: Sequence[Identifier], reverse: bool = False) -> Sequence[T]:
    """
    Sort `sequence` by the values of `keys`.

    >>> x = ['a', 'b', 'c']
    >>> keys = [3, 1, 2]
    >>> key_sort(x, keys=keys)
    ['b', 'c', 'a']
    """
    if len(sequence) != len(keys):
        raise ValueError(f"`sequence` and `keys` have different lengths: {len(sequence)=}; {len(keys)=}.")
    if len(sequence) == 0:
        return []
    zipped_sequence = list(zip(keys, sequence))
    sorted_zipped_sequence = sorted(zipped_sequence, key=lambda pair: pair[0], reverse=reverse)
    sorted_sequence = [pair[1] for pair in sorted_zipped_sequence]
    return sorted_sequence


def key_filter(sequence: Sequence[T], keys: Sequence[bool]) -> Sequence[T]:
    """
    Filter `sequence` by the values of `keys`.

    >>> x = ['a', 'b', 'c']
    >>> keys = [True, False, True]
    >>> key_filter(x, keys=keys)
    ['a', 'c']
    """
    if len(sequence) != len(keys):
        raise ValueError(f"`sequence` and `keys` have different lengths: {len(sequence)=}; {len(keys)=}.")
    return [element for element, key in zip(sequence, keys) if key]
    

def pack_intervals(intervals: Iterable[tuple[float, float]], max_offset:float=float('inf')) -> Sequence[int]:
    """
    Assign an non-negative offset to each input interval so that intervals sharing the same offset will not overlap with each other, while minimising offset values.
    Intervals are treated as being closed.

    >>> pack_intervals([(1, 2), (3, 4), (1, 3), (0, 5)])
    [0, 0, 1, 2]
    >>> pack_intervals([(1, 2), (1, 2), (1, 2), (3, 4)], max_offset=1)
    [0, 1, -1, 0]
    """
    occupied_intervals: list[list[tuple[float, float]]] = [[]]
    offsets = []
    for (start, end) in intervals:
        for offset, intervals in enumerate(occupied_intervals):
            for occupied_start, occupied_end in intervals:
                if (
                    occupied_start <= start <= occupied_end
                    or occupied_start <= end <= occupied_end
                    or start <= occupied_start <= end
                ):
                    break
            else:
                # Found place at the current offset.
                offsets.append(offset)
                occupied_intervals[offset].append((start, end))
                break
        else:
            # No places available.
            # Increment the offset.
            new_offset = offset + 1
            if new_offset > max_offset:
                # Reached max offset; try next interval
                offsets.append(-1) # -1 -> failed to pack within max offset
                continue
            occupied_intervals.append([(start, end)])
            offsets.append(offset + 1)
    return offsets
